"""click_demultiplex validators."""

from glob import glob
import os

from click_demultiplex import exceptions


def validate_patterns_are_files(patterns, check_size=True):
    """
    Check that a list of `patterns` are valid files.

    Arguments:
        patterns (list): a list of patterns to be check.
        check_size (bool): check size is not zero for all files matched.

    Returns:
        bool: True if all patterns match existing files.
    """
    for pattern in patterns:
        files = list(glob(pattern))

        if not files:
            msg = "{} pattern matched no files.".format(pattern)
            raise exceptions.ValidationError(msg)

        for i in files:
            if not os.path.isfile(i):
                msg = "{} is not a file.".format(i)
                raise exceptions.ValidationError(msg)

            if check_size and not os.path.getsize(i) > 0:
                msg = "{} is an empty file.".format(i)
                raise exceptions.ValidationError(msg)

    return True


def validate_patterns_are_dirs(patterns):
    """
    Check that a list of `patterns` are valid dirs.

    Arguments:
        patterns (list): a list of directory patterns.

    Returns:
        bool: True if all patterns match existing directories.
    """
    for pattern in patterns:
        dirs = list(glob(pattern))

        if not dirs:
            msg = "{} pattern matched no dirs.".format(pattern)
            raise exceptions.ValidationError(msg)

        for i in dirs:
            if not os.path.isdir(i):
                msg = "{} is not a directory.".format(i)
                raise exceptions.ValidationError(msg)

    return True



"""=click_demultiplex cli tests."""

from os.path import abspath
from os.path import dirname
from os.path import join
from glob import glob
import subprocess

from Bio import SeqIO
from click.testing import CliRunner
import click
import pandas as pd
import pytest

from click_demultiplex import cli
from click_demultiplex import commands


ROOT = abspath(dirname(__file__))
TEST_R1 = join(ROOT, 'data', 'test_R1.fastq')
TEST_R2 = join(ROOT, 'data', 'test_R2.fastq')
TEST_BARCODES = join(ROOT, 'data', 'test_barcodes.txt')


def test_cli(tmpdir):
    params = [
        "--r1", TEST_R1,
        "--r2", TEST_R2,
        "--outdir", tmpdir.strpath,
        "--barcodes", TEST_BARCODES,
    ]
    result = CliRunner().invoke(cli.main, params)
    assert result.exit_code == 0


def test_defaults(tmpdir):
    params = {
        'barcodes_path': TEST_BARCODES,
        'max_mismatches': 1,
        'not_trim': False,
        'output_dir': tmpdir.strpath,
        'overwrite': False,
        'prefix': '',
        'r1_path': TEST_R1,
        'r2_path': TEST_R2,
    }
    commands.demultiplex(*params.values())
    assert_output(**params)


def test_trim(tmpdir):
    params = {
        'barcodes_path': TEST_BARCODES,
        'max_mismatches': 1,
        'not_trim': True,
        'output_dir': tmpdir.strpath,
        'overwrite': True,
        'prefix': '',
        'r1_path': TEST_R1,
        'r2_path': TEST_R2,
    }
    commands.demultiplex(*params.values())
    assert_output(**params)


def test_prefix(tmpdir):
    params = {
        'barcodes_path': TEST_BARCODES,
        'max_mismatches': 1,
        'not_trim': True,
        'output_dir': tmpdir.strpath,
        'overwrite': True,
        'prefix': 'my_weird_prefix',
        'r1_path': TEST_R1,
        'r2_path': TEST_R2,
    }
    commands.demultiplex(*params.values())
    assert_output(**params)


def test_max_mismatches(tmpdir):
    params = {
        'barcodes_path': TEST_BARCODES,
        'max_mismatches': 0,
        'not_trim': True,
        'output_dir': tmpdir.strpath,
        'overwrite': True,
        'prefix': 'my_weird_prefix',
        'r1_path': TEST_R1,
        'r2_path': TEST_R2,
    }
    commands.demultiplex(*params.values())

    params['max_mismatches'] = 1
    commands.demultiplex(*params.values())
    assert_output(**params)

    params['max_mismatches'] = 2
    commands.demultiplex(*params.values())
    assert_output(**params)

    params['max_mismatches'] = 3
    commands.demultiplex(*params.values())
    assert_output(**params)


def test_overwrite(tmpdir):
    params = {
        'barcodes_path': TEST_BARCODES,
        'max_mismatches': 1,
        'not_trim': False,
        'output_dir': tmpdir.strpath,
        'overwrite': True,
        'prefix': '',
        'r1_path': TEST_R1,
        'r2_path': TEST_R2,
    }
    commands.demultiplex(*params.values())
    commands.demultiplex(*params.values())

    params['overwrite'] = False
    with pytest.raises(click.UsageError) as excinfo:
        commands.demultiplex(*params.values())
    assert 'pass --overwrite as an option' in excinfo.value.message


def assert_output(
        r1_path,
        r2_path,
        barcodes_path,
        output_dir,
        overwrite,
        prefix,
        not_trim,
        max_mismatches):

    # Parse r1 and r2
    multiplexed_r1 = SeqIO.parse(r1_path, 'fastq')
    original_sequence_length = len(next(multiplexed_r1))

    # Parse barcodes
    barcodes = commands.get_barcodes(barcodes_path)

    # Output files
    stats_file = join(output_dir, f'{prefix}result_stats.txt')
    output_files = glob(join(output_dir, '*.fastq'))

    # Parse Result stats file
    stats = pd.read_csv(
        stats_file,
        sep='\t',
        skiprows=1,
        skipfooter=1,
        engine='python',
        index_col=False
    )
    stats_dict = stats.set_index('Name').to_dict('index')

    # Assert number of output files
    assert len(barcodes.keys()) == len(stats)
    assert len(barcodes.keys()) * 2 == len(output_files)

    for name, barcode in barcodes.items():
        r1_filename = f'{prefix}{name}_R1.fastq'
        r2_filename = f'{prefix}{name}_R2.fastq'

        r1_path = join(output_dir, r1_filename)
        r2_path = join(output_dir, r2_filename)

        with open(r1_path, 'rt') as fr1, open(r2_path, 'rt') as fr2:
            records_r1 = list(SeqIO.parse(fr1, 'fastq'))
            records_r2 = list(SeqIO.parse(fr2, 'fastq'))

            assert len(records_r1) == len(records_r2)
            assert stats_dict[name]['Barcode'] == barcode
            assert stats_dict[name]['Count'] == len(records_r1)
            assert r1_filename in stats_dict[name]['Output R1 file']
            assert r2_filename in stats_dict[name]['Output R2 file']
            assert_quantity_of_filtered_sequences(stats_file, max_mismatches)

            for index in range(len(records_r1)):
                expected_sequence_length = original_sequence_length - (
                    0 if not_trim else len(barcode)
                )
                assert records_r1[index].id == records_r2[index].id
                assert len(records_r1[index].seq) == expected_sequence_length
                assert len(records_r2[index].seq) == expected_sequence_length


def assert_quantity_of_filtered_sequences(stats_file, max_mismatches):
    result_line = subprocess.check_output(['tail', '-1', stats_file])
    results_filtered_count = int(result_line.split()[3])
    expected_filtered_count = [35, 36, 60, 158, 250, 250, 250]
    assert expected_filtered_count[max_mismatches] == results_filtered_count



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?
You might be tempted to import things from __main__ later, but that will
cause problems, the code will get executed twice:

    - When you run `python -m click_demultiplex` python will execute
      `__main__.py` as a script. That means there won't be any
      `click_demultiplex.__main__` in `sys.modules`.

    - When you import __main__ it will get executed again (as a module) because
      there's no `click_demultiplex.__main__` in `sys.modules`.

Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import click

from click_demultiplex import __version__
from click_demultiplex import commands


@click.command()
@click.option(
    "--outdir",
    show_default=True,
    required=True,
    help="Path to output directory.",
    default=None)
@click.option(
    "--r1",
    required=True,
    type=click.Path(exists=True),
    help="Path to R1 fastq file. Reads in forward orientation")
@click.option(
    "--r2",
    required=True,
    type=click.Path(exists=True),
    help="Path to R2 fatsq file. Reads in reverse-complement orientation")
@click.option(
    "--barcodes",
    required=True,
    type=click.Path(exists=True),
    help="A text file with the barcodes in each line.")
@click.option(
    "--prefix",
    default="",
    help="String to add to output files.")
@click.option(
    "--no-trim",
    default=False,
    is_flag=True,
    help="Flag to avoid trimming the barcodes in each read.")
@click.option(
    "--overwrite",
    default=False,
    is_flag=True,
    help="Flag to overwrite the output files if they already exist.")
@click.option(
    "--max-mismatches",
    default=1,
    show_default=True,
    help="Maximum number of mismatches allowed in the barcode to demultiplex.")
@click.version_option(__version__)
def main(outdir, r1, r2, barcodes, no_trim, overwrite, prefix, max_mismatches):
    r"""
    Demultiplex a paired-end fastq file into several based on unique barcodes.

    The barcodes are sequences attached at the beginning of each read.
    By default, it trimms the barcodes off the demultiplexed reads,
    unless --no-trim is passed.

    The barcodes text file should be formatted to have 1 column with
    the barcodes, and an optional additional column to assign names to
    the demultiplexed result files.  the following structure:

        \b
        ATTCGT       A1
        ATATTC       A2
        TCGGAC       B1
        TCGAGG       B2
    """
    prefix = prefix if prefix.endswith('_') else f'{prefix}_'

    commands.demultiplex(
        barcodes_path=barcodes,
        max_mismatches=max_mismatches,
        no_trim=no_trim,
        prefix=prefix,
        output_dir=outdir,
        overwrite=overwrite,
        r1_path=r1,
        r2_path=r2,
    )

if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""click_demultiplex utils tests."""

from os.path import join
import os
import tarfile

from click_demultiplex import utils


def test_force_symlink(tmpdir):
    src = join(str(tmpdir), "src")
    dst = join(str(tmpdir), "dst")

    with open(src, "w") as f:
        f.write("Not empty.")

    utils.force_symlink(src, dst)
    assert os.path.islink(dst)


def test_force_symlink_overwrite(tmpdir):
    src = join(str(tmpdir), "src")
    dst = join(str(tmpdir), "dst")

    with open(src, "w") as f:
        f.write("Correct.")

    with open(dst, "w") as f:
        f.write("Wrong.")

    utils.force_symlink(src, dst)
    assert os.path.islink(dst)

    with open(dst, "r") as f:
        assert "Correct" in f.read()


def test_force_link(tmpdir):
    src = join(str(tmpdir), "src")
    dst = join(str(tmpdir), "dst")

    with open(src, "w") as f:
        f.write("Not empty.")

    utils.force_link(src, dst)
    assert os.path.isfile(dst)
    assert not os.path.islink(dst)


def test_force_link_overwrite(tmpdir):
    src = join(str(tmpdir), "src")
    dst = join(str(tmpdir), "dst")

    with open(src, "w") as f:
        f.write("Correct.")

    with open(dst, "w") as f:
        f.write("Wrong.")

    utils.force_link(src, dst)
    assert os.path.isfile(dst)
    assert not os.path.islink(dst)

    with open(dst, "r") as f:
        assert "Correct" in f.read()


def test_tar_dir(tmpdir):
    dst_dir = join(str(tmpdir), "dst_dir")
    source_dir = join(str(tmpdir), "source_dir")
    output_path = join(str(tmpdir), "source_dir.tar.gz")

    files = (
        (join(source_dir, "1"), "first file."),
        (join(source_dir, "2"), "second file."),
        (join(source_dir, "3"), "third file."),
        )

    os.makedirs(source_dir)
    os.makedirs(dst_dir)

    for i, j in files:
        with open(i, "w") as f:
            f.write(j)

    utils.tar_dir(output_path=output_path, source_dir=source_dir)
    assert tarfile.is_tarfile(output_path)

    # Remove files and dir
    for i in files:
        os.unlink(i[0])

    os.rmdir(source_dir)
    tar = tarfile.open(output_path)
    tar.extractall(path=dst_dir)
    tar.close()

    for i, j in files:
        i = i.replace(source_dir, join(dst_dir, "source_dir"))
        with open(i, "r") as f:
            assert j in f.read()



"""click_demultiplex main command."""

from os.path import join
import os
import gzip
import subprocess

from Bio import SeqIO
import click


# Utils for classifying Sequences
def hamming_distance(pattern1, pattern2):
    """Return the hamming distance between 2 patterns."""
    if len(pattern1) == len(pattern2):
        return sum([
            pattern1[index] != pattern2[index]
            for index in range(len(pattern1))
        ])
    raise Exception('Length of both reads do not match')


def find_best_barcode(record, barcodes):
    """Return the best hamming distance barcode."""
    distance = [
        hamming_distance(barcode, record.seq[:len(barcode)])
        for barcode in barcodes.values()
    ]
    min_distance = min(distance)
    matched_barcode = [
        barcode
        for barcode in barcodes.items()
    ][distance.index(min(distance))]
    return matched_barcode + (min_distance,)


def find_best_match(record1, record2, barcodes, max_mismatches):
    """Return the best match barcode between R1 and R2."""
    name1, barcode1, distance1 = find_best_barcode(record1, barcodes)
    name2, barcode2, distance2 = find_best_barcode(record2, barcodes)

    if distance1 <= distance2 and distance1 <= max_mismatches:
        return name1, barcode1
    elif distance2 <= distance1 and distance2 <= max_mismatches:
        return name2, barcode2
    return None, None


# Utils to manage files handles
def create_output_handle(output_dir, name, overwrite, prefix, direction):
    """Validate that output dir and files exist."""
    if not os.path.isdir(output_dir):
        raise click.UsageError(
            f'Output directory {output_dir} doesn`t exist.'
        )

    file_path = join(output_dir, f'{prefix}{name}_R{direction}.fastq')
    if os.path.isfile(file_path):
        if overwrite:
            os.remove(file_path)
        else:
            raise click.UsageError(message=(
                f'Output {file_path} file already exist. If you want to '
                'overwrite it, pass --overwrite as an option.'
            ))

    return open(file_path, 'a')


def get_files_handles(output_dir, barcodes, overwrite, prefix):
    """Manage file handles to keep files open during demultiplexing."""
    return {
        cell: {
            'r1': create_output_handle(output_dir, cell, overwrite, prefix, 1),
            'r2': create_output_handle(output_dir, cell, overwrite, prefix, 2)
        }
        for cell in barcodes.keys()
    }

def close_file_handles(file_handles):
    """Close every opened file."""
    for file_cell_handles in file_handles.values():
        file_cell_handles['r1'].close()
        file_cell_handles['r2'].close()


# Util to parse barcodes
def get_barcodes(barcodes_path):
    """
    Create a dictionary with the barcodes.

    Where the key is name that will be given to the output files, and
    the value is the barcode. If the text file have a 2nd column this
    value will be used as name, if not the same barcode will be the name.
    """
    barcodes_file = open(barcodes_path, 'r')
    barcodes_lines = [line.split() for line in barcodes_file.readlines()]
    return {
        (line[1] if len(line) == 2 else line[0]): line[0]
        for line in barcodes_lines
    }


# Utils to create output files
def create_output_stats(
        output_dir,
        stats,
        barcodes,
        output_handles,
        initial_count,
        prefix):
    """Print stats of results."""
    output_stats_path = join(output_dir, f'{prefix}result_stats.txt')

    with open(output_stats_path, 'w') as output_stats_file:
        output_stats_file.write('Stats of # of reads per barcode:\n\n')
        output_stats_file.write("{}\t{}\t{}\t{}\t{}\n".format(
            'Barcode', 'Name', 'Count', 'Output R1 file', 'Output R2 file'
        ))
        for name, count in stats.items():
            output_stats_file.write("{}\t{}\t{}\t{}\t{}\n".format(
                barcodes[name],
                name,
                count,
                output_handles[name]['r1'].name,
                output_handles[name]['r2'].name,
            ))
        final_count = sum(stats.values())
        output_stats_file.write(
            f'\nA total of {final_count} from {initial_count} '
            f'reads were demultiplexed.\n'
        )
    subprocess.check_call(["cat", output_stats_path])


# Main Function
def demultiplex(
        barcodes_path,
        max_mismatches,
        no_trim,
        output_dir,
        overwrite,
        prefix,
        r1_path,
        r2_path):
    """Demultiplex one file into several according the barcodes file."""
    # Parse barcode dictionary
    barcodes = get_barcodes(barcodes_path)

    # Get handles for output files
    output_handles = get_files_handles(output_dir, barcodes, overwrite, prefix)

    # Collect some stats
    records_by_cell = {key: 0 for key in barcodes.keys()}

    # Select open method depending of the format of the files
    open_r1 = gzip.open if r1_path.endswith('gz') else open
    open_r2 = gzip.open if r2_path.endswith('gz') else open

    # Open record one by one of each file, classify it and output to file.
    with open_r1(r1_path, 'rt') as fr1, open_r2(r2_path, 'rt') as fr2:

        # Get generators for both files
        records_r1_gen = SeqIO.parse(fr1, 'fastq')
        records_r2_gen = SeqIO.parse(fr2, 'fastq')

        label = f'Started demultiplexing files {r1_path} and {r2_path}'
        with click.progressbar(records_r1_gen, label=label) as bar:

            initial_count = 0

            for record_r1 in bar:
                record_r2 = next(records_r2_gen)

                # Store some stats
                initial_count += 1

                cell, barcode = find_best_match(
                    record_r1,
                    record_r2,
                    barcodes,
                    max_mismatches
                )

                if cell and barcode:

                    # Trim unless flag is passed
                    if not no_trim:
                        record_r1 = record_r1[len(barcode):]
                        record_r2 = record_r2[len(barcode):]

                    # Collect stats of demultiplexed ones
                    records_by_cell[cell] += 1

                    # Write to files
                    SeqIO.write(record_r1, output_handles[cell]['r1'], 'fastq')
                    SeqIO.write(record_r2, output_handles[cell]['r2'], 'fastq')

    # Close Handles
    close_file_handles(output_handles)

    # Print Stats
    create_output_stats(
        output_dir,
        records_by_cell,
        barcodes,
        output_handles,
        initial_count,
        prefix
    )



"""click_demultiplex specific exceptions."""


class PackageBaseException(Exception):

    """A base exception for click_demultiplex."""


class ValidationError(PackageBaseException):

    """A class to raise when a validation error occurs."""


class MissingRequirementError(PackageBaseException):

    """A class to raise when a requirement is missing."""


class MissingOutputError(PackageBaseException):

    """A class to raise when a file that should exist is missing."""


class ConfigurationError(PackageBaseException):

    """A class to raise when is not properly configured."""


class ImplementationError(PackageBaseException):

    """A class to raise when is not properly implemented."""


class CantBeRunError(PackageBaseException):

    """A class to raise when a pipeline just cannot be run."""


class MissingDataError(PackageBaseException):

    """A class to raise when data is missing."""



"""click_demultiplex utils."""

import os
import tarfile


def force_link(src, dst):
    """Force a link between src and dst."""
    try:
        os.unlink(dst)
        os.link(src, dst)
    except OSError:
        os.link(src, dst)


def force_symlink(src, dst):
    """Force a symlink between src and dst."""
    try:
        os.unlink(dst)
        os.symlink(src, dst)
    except OSError:
        os.symlink(src, dst)


def tar_dir(output_path, source_dir):
    """Compress a `source_dir` in `output_path`."""
    with tarfile.open(output_path, "w:gz") as tar:
        tar.add(source_dir, arcname=os.path.basename(source_dir))



"""click_demultiplex validators tests."""

from os.path import join
import pytest

from click_demultiplex import exceptions
from click_demultiplex import validators


def test_validate_patterns_are_files(tmpdir):
    """Create multiple files and test test_validate_patterns_are_files."""
    tmpdir_path = str(tmpdir)

    for i in range(5):
        with open(join(tmpdir_path, "empty" + str(i)), "w") as _:
            pass

    for i in range(11, 15):
        with open(join(tmpdir_path, "not_empty" + str(i)), "w") as f:
            f.write("I'm not empty.")

    empty = [join(tmpdir_path, "empty*")]
    not_empty = [join(tmpdir_path, "not_empty*")]
    none_existing = ["florentino", "ariza*"]
    none_file = [tmpdir_path]

    # Check empty files exist.
    assert validators.validate_patterns_are_files(empty, check_size=False)

    # Check files exist amd are not empty.
    assert validators.validate_patterns_are_files(not_empty, check_size=True)

    # Check that empty files raise error with default setting.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_files(empty)

    # Check that empty files raise error with flag.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_files(empty, check_size=True)

    # Check that pattern is not file.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_files(none_file, check_size=True)

    # Check that empty files raise error with flag.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_files(empty, check_size=True)

    # Check that invalid patterns raise ValidationError error.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_files(none_existing, check_size=True)


def test_validate_patterns_are_dirs(tmpdir):
    """test_validate_patterns_are_dirs."""
    tmpdir_path = str(tmpdir)
    file_patterns = [join(tmpdir_path, "a_file")]
    existing_patterns = [tmpdir_path]
    none_existing_patterns = ["florentino", "ariza*"]

    # Touch the file.
    with open(file_patterns[0], "w") as _:
        pass

    # Check empty files exist.
    assert validators.validate_patterns_are_dirs(existing_patterns)

    # Check that empty files raise error with default setting.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_dirs(none_existing_patterns)

    # Check that empty files raise error with flag.
    with pytest.raises(exceptions.ValidationError) as _:
        validators.validate_patterns_are_dirs(file_patterns)
